@@ -5,440 +5,12 @@
 #include <pybind11/stl.h>
 #include <cmath>
 #include <vector>
+#include <string>
 #include <boost/math/special_functions/gamma.hpp>
 
 
 namespace py = pybind11;
 
-int eulerian(int n, int m){
-    if (m >= n or n == 0){
-        return 0;
-    }
-    if (m == 0){
-        return 1;
-    }
-    return ((n - m) * eulerian(n - 1, m - 1) + (m + 1) * eulerian(n - 1, m));
-}
-
-std::vector<double> eulerian_all(int n){
-    int i = 0;
-    std::vector<double> res(n);
-    for(i = 0; i < n; i++){
-        res[i] = eulerian(n, i);
-    }
-    return res;
-}
-
-double log1mexp(double x){
-    if(x <= std::log(2.0)){
-        return std::log(-std::expm1(-x));
-    }else{
-        return std::log1p(-std::exp(-x));
-    }
-}
-
-
-std::vector<double> log1mexpvec(std::vector<double> x){
-    std::vector<double>::size_type i = 0;;
-    double eps = std::log(2.0);
-    std::vector<double> res(x.size());
-    for(i = 0; i < x.size(); i++){
-        if(x[i] <= eps){
-            res[i] = std::log(-std::expm1(-x[i]));
-        }else{
-            res[i] = std::log1p(-std::exp(-x[i]));
-        }
-    }
-    return res;
-}
-
-
-double polyneval(
-    std::vector<double> p,
-    double x
-){
-    std::vector<double>::size_type i = 0;
-    double res = 0.0;
-    for(i=0; i<p.size(); i++){
-        res = res * x + p[i];
-    }
-    return res;
-}
-
-
-std::vector<double> polyneval(
-    std::vector<double> p,
-    std::vector<double> x,
-    bool is_log_z = false
-){
-    std::vector<double>::size_type i = 0;
-    std::vector<double> res(x.size());
-    if(is_log_z){
-        for(i=0; i<x.size(); i++){
-            res[i] = polyneval(p, std::exp(x[i]));
-        }
-    }else{
-        for(i=0; i<x.size(); i++){
-            res[i] = polyneval(p, x[i]);
-        }
-    }
-    return res;
-}
-
-Eigen::VectorXd polyneval(
-    std::vector<double> p,
-    Eigen::VectorXd x,
-    bool is_log_z = false
-){
-    Eigen::EigenBase<Eigen::VectorXd>::Index i = 0;
-    Eigen::VectorXd res = Eigen::VectorXd(x.size());
-    if(is_log_z){
-        for(i=0; i<x.size(); i++){
-            res(i) = polyneval(p, std::exp(x(i)));
-        }
-    }else{
-        for(i=0; i<x.size(); i++){
-            res(i) = polyneval(p, x(i));
-        }
-    }
-    return res;
-}
-
-
-py::array_t<double> polyneval_py(
-    py::array_t<double> coef,
-    py::array_t<double> x,
-    bool is_log_z = false
-){
-    return py::cast(
-        polyneval(
-            coef.cast<std::vector<double>>(),
-            x.cast<Eigen::VectorXd>(),
-            is_log_z
-        )
-    );
-}
-
-
-std::vector<double> minus_vec(std::vector<double> x){
-    std::vector<double>::size_type i = 0;
-    std::vector<double> res(x.size());
-    for(i=0; i < x.size(); i++){
-        res[i] = -x[i];
-    }
-    return res;
-}
-
-
-std::vector<double> polylog(
-    std::vector<double> z,
-    int s,
-    bool is_log_z=false
-){
-    std::vector<double>::size_type i = 0;
-    std::vector<double> res = polyneval(
-        eulerian_all(-s),
-        z,
-        is_log_z
-    );
-    if(is_log_z){
-        for(i=0; i<z.size(); i++){
-            res[i] = std::log(res[i]) + \
-            z[i] - (-s + 1.0) * log1mexp(-z[i]);
-        }
-    }else{
-        for(i=0; i<z.size(); i++){
-            res[i] = std::log(res[i]) + \
-            std::log(z[i]) - (-s + 1.0) * std::log1p(-z[i]);
-        }
-    }
-    return res;
-}
-
-
-Eigen::VectorXd polylog(
-    Eigen::VectorXd z,
-    int s,
-    bool is_log_z=false
-){
-    Eigen::EigenBase<Eigen::VectorXd>::Index i = 0;
-    Eigen::VectorXd res = polyneval(
-        eulerian_all(-s),
-        z,
-        is_log_z
-    );
-    if(is_log_z){
-        for(i=0; i<z.size(); i++){
-            res(i) = std::log(res(i)) + \
-            z(i) - (-s + 1.0) * log1mexp(-z(i));
-        }
-    }else{
-        for(i=0; i<z.size(); i++){
-            res(i) = std::log(res(i)) + \
-            std::log(z(i)) - (-s + 1.0) * std::log1p(-z(i));
-        }
-    }
-    return res;
-}
-
-
-py::array_t<double> polylog_py(
-    py::array_t<double> z,
-    int s,
-    bool is_log_z=false
-){
-    return py::cast(
-        polylog(
-            z.cast<Eigen::VectorXd>(),
-            s,
-            is_log_z
-        )
-    );
-}
-
-
-Eigen::VectorXd pdf_frank(
-    Eigen::Matrix<
-        double,
-        Eigen::Dynamic,
-        Eigen::Dynamic> u_values,
-    double theta,
-    bool is_log = false){
-    Eigen::VectorXd dcopula;
-    if(theta == 0.0){
-        dcopula.setZero(
-            u_values.rows()
-        );
-        return dcopula;
-    }
-    int d;
-    double lp;
-    Eigen::VectorXd usum = Eigen::VectorXd(u_values.rows());
-    Eigen::VectorXd lu = Eigen::VectorXd(u_values.rows());
-    Eigen::VectorXd li = Eigen::VectorXd(u_values.rows());
-    Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> lpu = \
-        Eigen::MatrixXd(u_values.rows(), u_values.cols());
-    Eigen::VectorXd lpusum = Eigen::VectorXd(u_values.rows());
-    d = u_values.cols();
-    usum = u_values.rowwise().sum();
-    lp = log1mexp(theta);
-    lpu = (u_values.array() * theta).unaryExpr(&log1mexp).matrix();
-    lu = lpu.rowwise().sum();
-    lpu = (lpu.array() - lp).matrix();
-    lpusum = lpu.rowwise().sum();
-    lpusum = (lp + lpusum.array()).matrix();
-    li = polylog(
-        lpusum,
-        -(d - 1),
-        true
-    );
-    dcopula = (d - 1.0) * std::log(theta) + \
-    (li - (usum.array() * theta).matrix() - lu).array();
-    if(is_log){
-        return dcopula;
-    }
-    return dcopula.array().exp();
-}
-
-py::array_t<double> pdf_frank_py(
-    py::array_t<double> u_values,
-    double theta,
-    bool is_log = false
-){
-    return py::cast(
-        pdf_frank(
-            u_values.cast<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>>(),
-            theta,
-            is_log
-        )
-    );
-}
-
-
-Eigen::Matrix<
-        double,
-        Eigen::Dynamic,
-        Eigen::Dynamic> ipsi_clayton(
-    Eigen::Matrix<
-        double,
-        Eigen::Dynamic,
-        Eigen::Dynamic> u_values,
-    double theta,
-    bool is_log = false){
-    Eigen::Matrix<double,
-        Eigen::Dynamic,
-        Eigen::Dynamic> res;
-    res.setZero(
-            u_values.rows(),
-            u_values.cols()
-        );
-    double theta_sign = 1.0;
-    if(theta < 0.0){
-        theta_sign = -1.0;
-    }
-    res = theta_sign * ((u_values.array().pow(-theta)) - 1.0);
-    if(is_log){
-        return res.array().log();
-    }
-    return res;
-}
-
-
-py::array_t<double> ipsi_clayton_py(
-    py::array_t<double> u_values,
-    double theta,
-    bool is_log = false
-){
-    return py::cast(
-        ipsi_clayton(
-            u_values.cast<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>>(),
-            theta,
-            is_log
-        )
-    );
-}
-
-
-Eigen::VectorXd pdf_clayton(
-    Eigen::Matrix<
-        double,
-        Eigen::Dynamic,
-        Eigen::Dynamic> u_values,
-    double theta,
-    bool is_log = false){
-    if (theta == 0.0){
-        return Eigen::VectorXd::Zero(u_values.rows());
-    }
-    Eigen::VectorXd res(u_values.rows());
-    double d = float(u_values.cols());
-    Eigen::VectorXd lu = u_values.array().log().matrix().rowwise().sum();
-    Eigen::VectorXd t_var = ipsi_clayton(u_values, theta, false).rowwise().sum();
-    if(theta < 0.0){
-        for(auto i = 0; i < u_values.rows(); i++){
-            if(t_var(i) < 1.0){
-                res(i) = std::log1p(theta);
-                res(i) -= (1.0 + theta) * lu(i);
-                res(i) -= (d + 1.0 / theta) * std::log1p(-t_var(i));
-            }else{
-                res(i) = NAN;
-            }
-        }
-    }else{
-        res = (
-            theta * (
-                Eigen::VectorXd::LinSpaced(int(d) - 1, 1.0, d - 1.0)
-            ).array()
-        ).log1p();
-        res = res.matrix().sum() - (1.0 + theta) * lu.array() - (
-            d + 1.0 / theta
-        ) * t_var.array().log1p();
-    }
-    if(is_log){
-        return res;
-    }
-    return res.array().exp();
-}
-
-
-py::array_t<double> pdf_clayton_py(
-    py::array_t<double> u_values,
-    double theta,
-    bool is_log = false
-){
-    return py::cast(
-        pdf_clayton(
-            u_values.cast<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>>(),
-            theta,
-            is_log
-        )
-    );
-}
-
-
-Eigen::Matrix<
-        double,
-        Eigen::Dynamic,
-        Eigen::Dynamic> ipsi_gumbel(
-    Eigen::Matrix<
-        double,
-        Eigen::Dynamic,
-        Eigen::Dynamic> u_values,
-    double theta,
-    bool is_log = false){
-    if(is_log){
-        return theta * (-u_values.array().log()).log();
-    }
-    return (-u_values.array().log()).pow(theta);
-}
-
-
-py::array_t<double> ipsi_gumbel_py(
-    py::array_t<double> u_values,
-    double theta,
-    bool is_log = false
-){
-    return py::cast(
-        ipsi_gumbel(
-            u_values.cast<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>>(),
-            theta,
-            is_log
-        )
-    );
-}
-
-
-double fact(double n) {
-   if ((n==0.0)||(n==1.0))
-   return 1.0;
-   else
-   return n*fact(n-1.0);
-}
-
-
-int fact(int n) {
-   if ((n==0)||(n==1))
-   return 1;
-   else
-   return n*fact(n-1);
-}
-
-
-double log_poisscdf(double k, double lambda){
-    return std::log(boost::math::gamma_q(k + 1.0, lambda));
-}
-
-
-Eigen::MatrixXd lpoiss_check(
-    Eigen::VectorXd lx_var,
-    double alpha_var,
-    int d_var){
-    Eigen::VectorXd k = Eigen::VectorXd::LinSpaced(d_var, 1.0, double(d_var));
-    Eigen::VectorXd x = lx_var.array().exp();
-    Eigen::MatrixXd lppois = Eigen::MatrixXd::Zero(d_var, lx_var.size());
-    for(auto row = 0; row < d_var; row++){
-        for(auto col = 0; col < lx_var.size(); col++){
-            lppois(row, col) = log_poisscdf(double(d_var) - k(row), x(col));
-        }
-    }
-    return lppois;
-}
-
-
-py::array_t<double> lpoiss_check_py(
-    py::array_t<double> lx_var,
-    double alpha_var,
-    int d_var
-){
-    return py::cast(
-        lpoiss_check(
-            lx_var.cast<Eigen::Matrix<double, 1, Eigen::Dynamic>>(),
-            alpha_var,
-            d_var
-        )
-    );
-}
-
-
 Eigen::VectorXd lsum_row(
     Eigen::Matrix<
         double,
@@ -603,6 +175,553 @@ py::array_t<double> lssum_py(
 }
 
 
+
+int eulerian(int n, int m){
+    int dp[n + 1][m + 1];
+    memset(dp, 0, sizeof(dp));
+    // For each row from 1 to n
+    for (int i = 1; i <= n; i++) {
+        // For each column from 0 to m
+        for (int j = 0; j <= m; j++) {
+            // If i is greater than j
+            if (i > j) {
+                // If j is 0, then make that
+                // state as 1.
+                if (j == 0)
+                    dp[i][j] = 1;
+                // basic recurrence relation.
+                else
+                    dp[i][j] = ((i - j) *
+                     dp[i - 1][j - 1]) +
+                    ((j + 1) * dp[i - 1][j]);
+            }
+        }
+    }
+    return dp[n][m];
+}
+
+std::vector<double> eulerian_all(int n){
+    int i = 0;
+    std::vector<double> res(n);
+    for(i = 0; i < n; i++){
+        res[i] = eulerian(n, i);
+    }
+    return res;
+}
+
+double log1mexp(double x){
+    if(x <= std::log(2.0)){
+        return std::log(-std::expm1(-x));
+    }else{
+        return std::log1p(-std::exp(-x));
+    }
+}
+
+
+std::vector<double> log1mexpvec(std::vector<double> x){
+    std::vector<double>::size_type i = 0;;
+    double eps = std::log(2.0);
+    std::vector<double> res(x.size());
+    for(i = 0; i < x.size(); i++){
+        if(x[i] <= eps){
+            res[i] = std::log(-std::expm1(-x[i]));
+        }else{
+            res[i] = std::log1p(-std::exp(-x[i]));
+        }
+    }
+    return res;
+}
+
+
+double polyneval(
+    std::vector<double> p,
+    double x
+){
+    std::vector<double>::size_type i = 0;
+    double res = 0.0;
+    for(i=0; i<p.size(); i++){
+        res = res * x + p[i];
+    }
+    return res;
+}
+
+double polyneval(
+    Eigen::VectorXd p,
+    double x
+){
+    std::vector<double>::size_type i = 0;
+    double res = 0.0;
+    for(i=0; i<p.size(); i++){
+        res = res * x + p[i];
+    }
+    return res;
+}
+
+std::vector<double> polyneval(
+    std::vector<double> p,
+    std::vector<double> x,
+    bool is_log_z = false
+){
+    std::vector<double>::size_type i = 0;
+    std::vector<double> res(x.size());
+    if(is_log_z){
+        for(i=0; i<x.size(); i++){
+            res[i] = polyneval(p, std::exp(x[i]));
+        }
+    }else{
+        for(i=0; i<x.size(); i++){
+            res[i] = polyneval(p, x[i]);
+        }
+    }
+    return res;
+}
+
+Eigen::VectorXd polyneval(
+    std::vector<double> p,
+    Eigen::VectorXd x,
+    bool is_log_z = false
+){
+    Eigen::EigenBase<Eigen::VectorXd>::Index i = 0;
+    Eigen::VectorXd res = Eigen::VectorXd(x.size());
+    if(is_log_z){
+        for(i=0; i<x.size(); i++){
+            res(i) = polyneval(p, std::exp(x(i)));
+        }
+    }else{
+        for(i=0; i<x.size(); i++){
+            res(i) = polyneval(p, x(i));
+        }
+    }
+    return res;
+}
+
+Eigen::VectorXd polyneval(
+    Eigen::VectorXd p,
+    Eigen::VectorXd x,
+    bool is_log_z = false
+){
+    Eigen::EigenBase<Eigen::VectorXd>::Index i = 0;
+    Eigen::VectorXd res = Eigen::VectorXd(x.size());
+    if(is_log_z){
+        for(i=0; i<x.size(); i++){
+            res(i) = polyneval(p, std::exp(x(i)));
+        }
+    }else{
+        for(i=0; i<x.size(); i++){
+            res(i) = polyneval(p, x(i));
+        }
+    }
+    return res;
+}
+
+
+py::array_t<double> polyneval_py(
+    py::array_t<double> coef,
+    py::array_t<double> x,
+    bool is_log_z = false
+){
+    return py::cast(
+        polyneval(
+            coef.cast<std::vector<double>>(),
+            x.cast<Eigen::VectorXd>(),
+            is_log_z
+        )
+    );
+}
+
+
+std::vector<double> minus_vec(std::vector<double> x){
+    std::vector<double>::size_type i = 0;
+    std::vector<double> res(x.size());
+    for(i=0; i < x.size(); i++){
+        res[i] = -x[i];
+    }
+    return res;
+}
+
+
+std::vector<double> polylog(
+    std::vector<double> z,
+    int s,
+    bool is_log_z=false
+){
+    std::vector<double>::size_type i = 0;
+    std::vector<double> res = polyneval(
+        eulerian_all(-s),
+        z,
+        is_log_z
+    );
+    if(is_log_z){
+        for(i=0; i<z.size(); i++){
+            res[i] = std::log(res[i]) + \
+            z[i] - (-s + 1.0) * log1mexp(-z[i]);
+        }
+    }else{
+        for(i=0; i<z.size(); i++){
+            res[i] = std::log(res[i]) + \
+            std::log(z[i]) - (-s + 1.0) * std::log1p(-z[i]);
+        }
+    }
+    return res;
+}
+
+
+Eigen::VectorXd polylog(
+    Eigen::VectorXd z,
+    int s,
+    bool is_log_z=false
+){
+    Eigen::EigenBase<Eigen::VectorXd>::Index i = 0;
+    Eigen::VectorXd res = polyneval(
+        eulerian_all(-s),
+        z,
+        is_log_z
+    );
+    if(is_log_z){
+        for(i=0; i<z.size(); i++){
+            res(i) = std::log(res(i)) + \
+            z(i) - (-s + 1.0) * log1mexp(-z(i));
+        }
+    }else{
+        for(i=0; i<z.size(); i++){
+            res(i) = std::log(res(i)) + \
+            std::log(z(i)) - (-s + 1.0) * std::log1p(-z(i));
+        }
+    }
+    return res;
+}
+
+Eigen::VectorXd polylog(
+    Eigen::VectorXd z,
+    int s,
+    Eigen::VectorXd eul,
+    bool is_log_z=false
+){
+    Eigen::EigenBase<Eigen::VectorXd>::Index i = 0;
+    Eigen::VectorXd res = polyneval(
+        eul,
+        z,
+        is_log_z
+    );
+    if(is_log_z){
+        for(i=0; i<z.size(); i++){
+            res(i) = std::log(res(i)) + \
+            z(i) - (-s + 1.0) * log1mexp(-z(i));
+        }
+    }else{
+        for(i=0; i<z.size(); i++){
+            res(i) = std::log(res(i)) + \
+            std::log(z(i)) - (-s + 1.0) * std::log1p(-z(i));
+        }
+    }
+    return res;
+}
+
+
+
+py::array_t<double> polylog_py(
+    py::array_t<double> z,
+    int s,
+    bool is_log_z=false
+){
+    return py::cast(
+        polylog(
+            z.cast<Eigen::VectorXd>(),
+            s,
+            is_log_z
+        )
+    );
+}
+
+
+Eigen::VectorXd pdf_frank(
+    Eigen::Matrix<
+        double,
+        Eigen::Dynamic,
+        Eigen::Dynamic> u_values,
+    double theta,
+    bool is_log = false){
+    Eigen::VectorXd dcopula;
+    if(theta == 0.0){
+        dcopula.setZero(
+            u_values.rows()
+        );
+        return dcopula;
+    }
+    int d;
+    double lp;
+    Eigen::VectorXd usum = Eigen::VectorXd(u_values.rows());
+    Eigen::VectorXd lu = Eigen::VectorXd(u_values.rows());
+    Eigen::VectorXd li = Eigen::VectorXd(u_values.rows());
+    Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> lpu = \
+        Eigen::MatrixXd(u_values.rows(), u_values.cols());
+    Eigen::VectorXd lpusum = Eigen::VectorXd(u_values.rows());
+    d = u_values.cols();
+    usum = u_values.rowwise().sum();
+    lp = log1mexp(theta);
+    lpu = (u_values.array() * theta).unaryExpr(&log1mexp).matrix();
+    lu = lpu.rowwise().sum();
+    lpu = (lpu.array() - lp).matrix();
+    lpusum = lpu.rowwise().sum();
+    lpusum = (lp + lpusum.array()).matrix();
+    std::vector<double> eul =  eulerian_all(d - 1);
+    li = polylog(
+        lpusum,
+        -(d - 1),
+        true
+    );
+    dcopula = (d - 1.0) * std::log(theta) + \
+    (li - (usum.array() * theta).matrix() - lu).array();
+    if(is_log){
+        return dcopula;
+    }
+    return dcopula.array().exp();
+}
+
+Eigen::VectorXd pdf_frank(
+    Eigen::Matrix<
+        double,
+        Eigen::Dynamic,
+        Eigen::Dynamic> u_values,
+    double theta,
+    Eigen::VectorXd eul,
+    bool is_log = false){
+    Eigen::VectorXd dcopula;
+    if(theta == 0.0){
+        dcopula.setZero(
+            u_values.rows()
+        );
+        return dcopula;
+    }
+    int d;
+    double lp;
+    Eigen::VectorXd usum = Eigen::VectorXd(u_values.rows());
+    Eigen::VectorXd lu = Eigen::VectorXd(u_values.rows());
+    Eigen::VectorXd li = Eigen::VectorXd(u_values.rows());
+    Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> lpu = \
+        Eigen::MatrixXd(u_values.rows(), u_values.cols());
+    Eigen::VectorXd lpusum = Eigen::VectorXd(u_values.rows());
+    d = u_values.cols();
+    usum = u_values.rowwise().sum();
+    lp = log1mexp(theta);
+    lpu = (u_values.array() * theta).unaryExpr(&log1mexp).matrix();
+    lu = lpu.rowwise().sum();
+    lpu = (lpu.array() - lp).matrix();
+    lpusum = lpu.rowwise().sum();
+    lpusum = (lp + lpusum.array()).matrix();
+    li = polylog(
+        lpusum,
+        -(d - 1),
+        eul,
+        true
+    );
+    dcopula = (d - 1.0) * std::log(theta) + \
+    (li - (usum.array() * theta).matrix() - lu).array();
+    if(is_log){
+        return dcopula;
+    }
+    return dcopula.array().exp();
+}
+
+py::array_t<double> pdf_frank_py(
+    py::array_t<double> u_values,
+    double theta,
+    py::array_t<double> eul,
+    bool is_log = false){
+    return py::cast(
+        pdf_frank(
+            u_values.cast<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>>(),
+            theta,
+            eul.cast<Eigen::VectorXd>(),
+            is_log
+        )
+    );
+}
+
+
+Eigen::Matrix<
+        double,
+        Eigen::Dynamic,
+        Eigen::Dynamic> ipsi_clayton(
+    Eigen::Matrix<
+        double,
+        Eigen::Dynamic,
+        Eigen::Dynamic> u_values,
+    double theta,
+    bool is_log = false){
+    Eigen::Matrix<double,
+        Eigen::Dynamic,
+        Eigen::Dynamic> res;
+    res.setZero(
+            u_values.rows(),
+            u_values.cols()
+        );
+    double theta_sign = 1.0;
+    if(theta < 0.0){
+        theta_sign = -1.0;
+    }
+    res = theta_sign * ((u_values.array().pow(-theta)) - 1.0);
+    if(is_log){
+        return res.array().log();
+    }
+    return res;
+}
+
+
+py::array_t<double> ipsi_clayton_py(
+    py::array_t<double> u_values,
+    double theta,
+    bool is_log = false
+){
+    return py::cast(
+        ipsi_clayton(
+            u_values.cast<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>>(),
+            theta,
+            is_log
+        )
+    );
+}
+
+
+Eigen::VectorXd pdf_clayton(
+    Eigen::Matrix<
+        double,
+        Eigen::Dynamic,
+        Eigen::Dynamic> u_values,
+    double theta,
+    bool is_log = false){
+    if (theta == 0.0){
+        return Eigen::VectorXd::Zero(u_values.rows());
+    }
+    Eigen::VectorXd res(u_values.rows());
+    double d = float(u_values.cols());
+    Eigen::VectorXd lu = u_values.array().log().matrix().rowwise().sum();
+    Eigen::VectorXd t_var = ipsi_clayton(u_values, theta, false).rowwise().sum();
+    if(theta < 0.0){
+        for(auto i = 0; i < u_values.rows(); i++){
+            if(t_var(i) < 1.0){
+                res(i) = std::log1p(theta);
+                res(i) -= (1.0 + theta) * lu(i);
+                res(i) -= (d + 1.0 / theta) * std::log1p(-t_var(i));
+            }else{
+                res(i) = NAN;
+            }
+        }
+    }else{
+        res = (
+            theta * (
+                Eigen::VectorXd::LinSpaced(int(d) - 1, 1.0, d - 1.0)
+            ).array()
+        ).log1p();
+        res = res.matrix().sum() - (1.0 + theta) * lu.array() - (
+            d + 1.0 / theta
+        ) * t_var.array().log1p();
+    }
+    if(is_log){
+        return res;
+    }
+    return res.array().exp();
+}
+
+
+py::array_t<double> pdf_clayton_py(
+    py::array_t<double> u_values,
+    double theta,
+    bool is_log = false){
+    return py::cast(
+        pdf_clayton(
+            u_values.cast<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>>(),
+            theta,
+            is_log
+        )
+    );
+}
+
+
+Eigen::Matrix<
+        double,
+        Eigen::Dynamic,
+        Eigen::Dynamic> ipsi_gumbel(
+    Eigen::Matrix<
+        double,
+        Eigen::Dynamic,
+        Eigen::Dynamic> u_values,
+    double theta,
+    bool is_log = false){
+    if(is_log){
+        return theta * (-u_values.array().log()).log();
+    }
+    return (-u_values.array().log()).pow(theta);
+}
+
+
+py::array_t<double> ipsi_gumbel_py(
+    py::array_t<double> u_values,
+    double theta,
+    bool is_log = false
+){
+    return py::cast(
+        ipsi_gumbel(
+            u_values.cast<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>>(),
+            theta,
+            is_log
+        )
+    );
+}
+
+
+double fact(double n) {
+   if ((n==0.0)||(n==1.0))
+   return 1.0;
+   else
+   return n*fact(n-1.0);
+}
+
+
+int fact(int n) {
+   if ((n==0)||(n==1))
+   return 1;
+   else
+   return n*fact(n-1);
+}
+
+
+double log_poisscdf(double k, double lambda){
+    return std::log(boost::math::gamma_q(k + 1.0, lambda));
+}
+
+
+Eigen::MatrixXd lpoiss_check(
+    Eigen::VectorXd lx_var,
+    double alpha_var,
+    int d_var){
+    Eigen::VectorXd k = Eigen::VectorXd::LinSpaced(d_var, 1.0, double(d_var));
+    Eigen::VectorXd x = lx_var.array().exp();
+    Eigen::MatrixXd lppois = Eigen::MatrixXd::Zero(d_var, lx_var.size());
+    for(auto row = 0; row < d_var; row++){
+        for(auto col = 0; col < lx_var.size(); col++){
+            lppois(row, col) = log_poisscdf(double(d_var) - k(row), x(col));
+        }
+    }
+    return lppois;
+}
+
+
+py::array_t<double> lpoiss_check_py(
+    py::array_t<double> lx_var,
+    double alpha_var,
+    int d_var
+){
+    return py::cast(
+        lpoiss_check(
+            lx_var.cast<Eigen::Matrix<double, 1, Eigen::Dynamic>>(),
+            alpha_var,
+            d_var
+        )
+    );
+}
+
+
 Eigen::VectorXd signff(
     double alpha,
     Eigen::VectorXd j,
@@ -738,8 +857,7 @@ Eigen::VectorXd pdf_gumbel(
 py::array_t<double> pdf_gumbel_py(
     py::array_t<double> u_values,
     double theta,
-    bool is_log = false
-){
+    bool is_log = false){
     return py::cast(
         pdf_gumbel(
             u_values.cast<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>>(),
@@ -822,10 +940,11 @@ PYBIND11_MODULE(c_archimedean, m) {
     m.def("pdf_frank", &pdf_frank_py, R"pbdoc(
         :param u_values: np.array
         :param theta: double
+        :param eul: np.array
         :param is_log: bool
         :return:
     )pbdoc",
-    py::arg("u_values"), py::arg("theta"), py::arg("is_log"));
+    py::arg("u_values"), py::arg("theta"), py::arg("eul"), py::arg("is_log"));
 
     m.def("ipsi_clayton", &ipsi_clayton_py, R"pbdoc(
         :param u_values: np.array
